use std::fs;

const GEAR:char = '*';

#[derive(Default,Debug,Clone,Copy,PartialEq)]
struct Gear{
    pub coords:Coords,
    pub side_one:usize,
    pub side_two:usize,
}

impl Gear{
    pub fn same_location(&self,coords:Coords) -> bool{
        return self.coords.x == coords.x && self.coords.y == coords.y
    }
    pub fn add_value(&mut self, side:usize) -> usize{
        if self.side_one == 0{
            self.side_one = side;
        } else if self.side_two == 0{
            self.side_two = side;
        };
        self.side_one * self.side_two
    }
}

#[derive(Default,Debug,Clone,Copy,PartialEq)]
struct Coords{
    pub x:usize,
    pub y:usize
}

const FILE_NAME:&str = "./data/data.txt";
fn main() {
    let file_contents = fs::read_to_string(FILE_NAME).unwrap();
    let lines:Vec<&str> = file_contents.split('\n').collect();
    let mut final_output = 0_usize;
    let mut map:Vec<Vec<char>> = Vec::new();
    for line in lines{
        let line_string = line.trim().to_string();
        map.push(line_string.chars().collect());
    }
    let mut all_gears:Vec<Gear> = Vec::new();
    for i in 0..(map.len()){
        let line = map.get(i).unwrap();
        for j in (0..(line.len())).rev(){
            let current_char = *line.get(j).unwrap();
            if current_char == GEAR{
                all_gears.push( Gear{coords:Coords{ y:i,x:j},side_one:0,side_two:0} );
            }
        };
    };
    for i in 0..(map.len()){
        let mut place = 0;
        let mut next_number = 0usize;
        let mut continue_adding = false;
        let mut back_index = 0;
        let line = map.get(i).unwrap();
        for j in (0..(line.len())).rev(){
            let current_char = *line.get(j).unwrap();
            if let Some(digit) = current_char.to_digit(10){
                if is_back(line.clone(), j){
                    next_number += digit as usize * 10_usize.pow(place);
                    place += 1;
                    continue_adding = true;
                    back_index = j;
                } else if continue_adding{
                    next_number += digit as usize * 10_usize.pow(place);
                    place += 1;
                };
                if is_front(line.clone(),j){
                    if let Some(gear_location) = is_a_part_number(map.clone(),back_index,j,i){
                        for gear in &mut all_gears{
                            if gear.same_location(gear_location){
                                final_output += gear.add_value(next_number);
                            }
                        }
                    } 
                    next_number = 0;
                    place = 0;
                    continue_adding = false;
                }
            };
        };
    };
    println!("{}",final_output);
}

fn is_a_part_number(map:Vec<Vec<char>>,back_index:usize,front_index:usize,row:usize) -> Option<Coords>{
    if front_index > 0{
        if is_special_char(*map.get(row).unwrap().get(front_index - 1).unwrap()){
            return Some(Coords{y:row,x:(front_index-1)});
        }
    }
    if let Some(location) = check_above(map.clone(), front_index, back_index, row){
        return Some(location);
    }
    if let Some(location) = check_below(map.clone(), front_index, back_index, row){
        return Some(location);
    }
    if let Some(back_check) = map.get(row).unwrap().get(back_index + 1){
        if is_special_char(*back_check){
            return Some(Coords{y:row,x:(back_index+1)});
        }
    }
    None
}

fn is_back(line:Vec<char>,index:usize) -> bool{
    line.get(index + 1).is_some_and(|x| !x.is_digit(10)) || line.get(index + 1).is_none()
}

fn is_front(line:Vec<char>,index:usize) -> bool{
    if index == 0 {
        true
    } else { line.get(index - 1).is_some_and(|x| !x.is_digit(10)) }
}

fn check_above(map:Vec<Vec<char>>,front_index:usize,back_index:usize, current_row:usize) -> Option<Coords>{
    if current_row == 0 { return None; }
    let line = map.get(current_row - 1).unwrap();
    let min_val:usize;
    if front_index == 0{
        min_val = 0;
    } else {
        min_val = front_index - 1;
    }
    for i in min_val..=(back_index+1){
        if line.get(i).is_some_and(|x| is_special_char(*x)){
            return Some(Coords{y:current_row-1,x:i});
        }
    }
    None
}

fn is_special_char(char:char) -> bool{ char == GEAR }

fn check_below(map:Vec<Vec<char>>,front_index:usize,back_index:usize, current_row:usize) -> Option<Coords>{
    if let Some(line) = map.get(current_row + 1){
        let min_val:usize;
        if front_index == 0{
            min_val = 0;
        } else {
            min_val = front_index - 1;
        }
        for i in min_val..=(back_index+1){
            if line.get(i).is_some_and(|x| is_special_char(*x)){
                return Some(Coords{y:current_row+1,x:i});
            }
        }
    }
    None
}
